import pandas as pd
import os
import pathlib
from path_plot_config import data_dir, int_dir, out_dir

# Set up table-specific directories
tab_dir = out_dir / 'Summary Statistics'



# ---------------------------------------------------------------------------------------- #
# For the annual data
# ---------------------------------------------------------------------------------------- #

# ## Creating tables of statistics

# Importing cleaned Compustat data

df_cpi = pd.read_csv(int_dir / "main_annual.csv")

# Normalizing PPI with CPI
df_cpi['PPI_CPI'] = df_cpi['ppi']/df_cpi['CPI']*100

main = df_cpi[['gvkey','year','cogs','sale','ind2d','firm_level_markup','ppi','CPI','ind2d_definition']].copy()

main.rename(columns={'sale':'sale_raw',
                   'ppi':'PPI'}, inplace=True)
main = main[main.year != 2022]
main.reset_index(drop=True, inplace=True)

# CPI inflation adjusting the sales revenue
main['sale'] = (main['sale_raw']/main['CPI'])*100

# Editing the 2-digit industry codes
main['ind2d'] = main['ind2d'].astype(str)

# Manufacturing
main.loc[main.ind2d == '31', 'ind2d'] = '31-33'
main.loc[main.ind2d == '32', 'ind2d'] = '31-33'
main.loc[main.ind2d == '33', 'ind2d'] = '31-33'

# Retail trade
main.loc[main.ind2d == '44', 'ind2d'] = '44-45'
main.loc[main.ind2d == '45', 'ind2d'] = '44-45'

# Transportation and warehousing
main.loc[main.ind2d == '48', 'ind2d'] = '48-49'
main.loc[main.ind2d == '49', 'ind2d'] = '48-49'


# Adjusting the sample

# Length of firm data truncation
df_y = (main.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d','ind2d_definition']]
        .agg(['first','last'])
        .reset_index())
df_y['periods'] = df_y['year','last'] - df_y['year','first'] 
df_y.columns = df_y.columns.map(','.join).str.strip(',')
df_y = df_y[['gvkey','periods']]

# Combine length of firms data into the main dataframe
df = pd.merge(main,df_y,on='gvkey')
df = df[df.periods >= 4]

# Choosing the years
df = df[df.year >= 1980]
df = df[df.year <= 2018]
df.reset_index(drop=True, inplace=True)


# ## Data for all firms

# Column (i) Number of all firms by 2-digit industries

firms = df[['ind2d','gvkey']]
firms = firms.drop_duplicates(subset = 'gvkey')

n_firms = firms.pivot_table(columns=['ind2d'], aggfunc='size')
n_firms = pd.DataFrame(n_firms)                          # To check the number of observations
n_firms.reset_index(inplace=True)
n_firms.rename(columns={0:'N_firms'}, inplace=True) 

# Total number of firms
print('Total number of firms is',n_firms.N_firms.sum())


# For Column (iii) summing up revenue for all firms

df_revenue = df.groupby('ind2d').sum(numeric_only = True)
df_revenue = df_revenue['sale']
df_revenue = pd.DataFrame(df_revenue)
df_revenue.reset_index(inplace=True)
df_revenue

col123 = pd.merge(n_firms, df_revenue, on='ind2d')


# ## Data for matched firms

# For Column (ii) share of all firms matched to PPI data

df_matched = main.dropna(subset=['PPI']).copy()

df_matched.rename(columns={'cogs_D':'cogs',
                    'sale_D':'sale',
                   'ppi':'PPI'}, inplace=True)

# Adjusting the sample
# Length of firm data truncation
df_y = (df_matched.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d','ind2d_definition']]
        .agg(['first','last'])
        .reset_index())
df_y['periods'] = df_y['year','last'] - df_y['year','first'] 
df_y.columns = df_y.columns.map(','.join).str.strip(',')
df_y = df_y[['gvkey','periods']]

# Combine length of firms data into the main dataframe
df_matched = pd.merge(df_matched,df_y,on='gvkey')
df_matched = df_matched[df_matched.periods >= 4]

# Choosing the years
df_matched = df_matched[df_matched.year >= 1980]
df_matched = df_matched[df_matched.year <= 2018]
df_matched.reset_index(drop=True, inplace=True)


# Number of firms per 2-digit industry

firms_matched = df_matched[['ind2d','gvkey']]
firms_matched = firms_matched.drop_duplicates(subset = 'gvkey')

n_firms_matched = firms_matched.pivot_table(columns=['ind2d'], aggfunc='size')
n_firms_matched = pd.DataFrame(n_firms_matched)                          # To check the number of observations
n_firms_matched.reset_index(inplace=True)
n_firms_matched.rename(columns={0:'N_firms_m'}, inplace=True) 

# Number of matched firms
print('Total number of firms matched:',n_firms_matched.N_firms_m.sum())


# Revenue share of the matched firms

df_revenue_m = df_matched.groupby('ind2d').sum(numeric_only = True)
df_revenue_m = df_revenue_m['sale']
df_revenue_m = pd.DataFrame(df_revenue_m)
df_revenue_m.reset_index(inplace=True)
df_revenue_m.rename(columns={'sale':'sale_m'}, inplace=True) 

# Adding columns to the main table
col123_m = pd.merge(n_firms_matched,df_revenue_m,on='ind2d')
col123_f = pd.merge(col123,col123_m,on='ind2d',how ='outer')
col123_f  = col123_f.fillna(0)


# Calculating totals and adding as a row

col123_t = col123_f.sum(numeric_only=True)
col123_t = pd.DataFrame(col123_t).T
col123_t = pd.concat([col123_t,col123_f])
col123_t.reset_index(drop=True, inplace=True)


# Calculating values for Columns ii-iii

col123_t['share_firms_matched'] = col123_t['N_firms_m'] / col123_t['N_firms']
col123_t['share_revenue_matched'] = col123_t['sale_m'] / col123_t['sale']

col123_t = col123_t[['ind2d','N_firms','share_firms_matched','share_revenue_matched']].copy()
col123_t.reset_index(drop=True, inplace=True)


# ## Adding column (iv)

# Column (iv) number of 6-digit codes in matched data

# Importing main dataset to get 6-digit industries
naics = pd.read_csv(int_dir / "main_annual.csv", usecols = ['gvkey','naics'])
naics = naics.drop_duplicates(subset = 'gvkey')

# Merging with the data for matched firms
df_matched = pd.merge(df_matched,naics,on='gvkey')
df_matched['naics'] = df_matched['naics'].astype(int)
df_matched['naics'] = df_matched['naics'].astype(str)

df_6digit = df_matched[['ind2d','naics']]
df_6digit = df_6digit.drop_duplicates(subset = 'naics')
df_6digit=df_6digit[df_6digit.naics.apply(lambda x: len(str(x))==6)]

n_6digit = df_6digit.pivot_table(columns=['ind2d'], aggfunc='size')
n_6digit = pd.DataFrame(n_6digit)                          
n_6digit.reset_index(inplace=True)
n_6digit.rename(columns={0:'N_6d_codes'}, inplace=True) 

stat_table = pd.merge(col123_t,n_6digit,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)
stat_table['N_6d_codes'] = stat_table['N_6d_codes'].astype(int)


# ### Adding column (v)

# Column (v) average length of firm data

df_years = (df_matched.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d','ind2d_definition']]
        .agg(['first','last'])
        .reset_index())
df_years['periods'] = df_years['year','last'] - df_years['year','first'] + 1
df_years.columns = df_years.columns.map(','.join).str.strip(',')
df_years = df_years[['ind2d,first','periods']]
df_years.rename(columns={'ind2d,first':'ind2d'}, inplace=True)
mean_periods = df_years['periods'].mean()

df_n_years = df_years.groupby('ind2d').mean(numeric_only = True)
stat_table = pd.merge(stat_table,df_n_years,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)

# Adding industry definitions

defs = df[['ind2d','ind2d_definition']]
defs = defs.drop_duplicates(subset = 'ind2d')

stat_table = pd.merge(stat_table,defs,on='ind2d',how='outer')

stat_table.loc[0,'ind2d_definition'] = 'Total'
stat_table.loc[0,'N_6d_codes'] = stat_table.N_6d_codes[1:].sum()
stat_table.loc[0,'periods'] = mean_periods
stat_table.loc[0,'ind2d'] = 'Total'

# Formatting the table

stat_table_annual = stat_table[['ind2d','ind2d_definition','N_firms',
                         'share_firms_matched','share_revenue_matched','N_6d_codes','periods']]
stat_table_annual.rename(columns={'ind2d':'Industry',
                          'ind2d_definition':'Definition',
                          'N_firms':'Number of firms',
                          'share_firms_matched': 'Share of firms matched',
                          'N_6d_codes':'Number of 6-digit industries',
                          'periods':'Number of periods (years)'},inplace=True)



# Exporting the table
stat_table_annual.round(4).to_csv(tab_dir / 'Stats Table Annual.csv')

# Saving as Tex file
with open(tab_dir / 'Stats Table Annual.tex','w') as tf:
    tf.write(stat_table_annual
             .rename(columns ={'share_revenue_matched': 'Share of revenue matched'})
             .round(2)
             .to_latex(index=False))


# ## Generating table of statistics (with truncation)

# For Column (ii) share of all firms matched to PPI data

df_matched = main.dropna(subset=['PPI']).copy()

df_matched.rename(columns={'cogs_D':'cogs',
                    'sale_D':'sale',
                   'ppi':'PPI'}, inplace=True)

# Adjusting the sample
# Length of firm data truncation
df_y = (df_matched.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d','ind2d_definition']]
        .agg(['first','last'])
        .reset_index())
df_y['periods'] = df_y['year','last'] - df_y['year','first'] 
df_y.columns = df_y.columns.map(','.join).str.strip(',')
df_y = df_y[['gvkey','periods']]

# Combine length of firms data into the main dataframe
df_matched = pd.merge(df_matched,df_y,on='gvkey')
df_matched = df_matched[df_matched.periods >= 4]

# Choosing the years
df_matched = df_matched[df_matched.year >= 1980]
df_matched = df_matched[df_matched.year <= 2018]
df_matched.reset_index(drop=True, inplace=True)

# Importing CAGR data for truncation (following Figure 2)

cagr = pd.read_csv(int_dir / "scatter_annual.csv",usecols=['gvkey','cagr_PPI','cagr_markup'])
cagr = cagr.drop_duplicates()

df_matched = pd.merge(df_matched,cagr,on='gvkey')

# Applying Figure 2 truncations
# Truncating the data
df_matched = df_matched[df_matched.cagr_PPI <= 20]
df_matched = df_matched[df_matched.cagr_PPI >= -20]
df_matched = df_matched[df_matched.cagr_markup <= 100]

# Number of firms per 2-digit industry

firms_matched = df_matched[['ind2d','gvkey']]
firms_matched = firms_matched.drop_duplicates(subset = 'gvkey')

n_firms_matched = firms_matched.pivot_table(columns=['ind2d'], aggfunc='size')
n_firms_matched = pd.DataFrame(n_firms_matched)                          
n_firms_matched.reset_index(inplace=True)
n_firms_matched.rename(columns={0:'N_firms_m'}, inplace=True) 

# Number of matched firms
print('Number of matched firms after truncation:',n_firms_matched.N_firms_m.sum())


# Revenue share of the matched firms

df_revenue_m = df_matched.groupby('ind2d').sum(numeric_only = True)
df_revenue_m = df_revenue_m['sale']
df_revenue_m = pd.DataFrame(df_revenue_m)
df_revenue_m.reset_index(inplace=True)
df_revenue_m.rename(columns={'sale':'sale_m'}, inplace=True) 

# Adding columns to the main table

col123_m = pd.merge(n_firms_matched,df_revenue_m,on='ind2d')
col123_f = pd.merge(col123,col123_m,on='ind2d',how ='outer')
col123_f  = col123_f.fillna(0)

# Calculating totals and adding as a row

col123_t = col123_f.sum(numeric_only=True)
col123_t = pd.DataFrame(col123_t).T
col123_t = pd.concat([col123_t,col123_f])
col123_t.reset_index(drop=True, inplace=True)

# Calculating values for Columns ii-iii

col123_t['share_firms_matched'] = col123_t['N_firms_m'] / col123_t['N_firms']
col123_t['share_revenue_matched'] = col123_t['sale_m'] / col123_t['sale']

col123_t = col123_t[['ind2d','N_firms','share_firms_matched','share_revenue_matched']].copy()
col123_t.reset_index(drop=True, inplace=True)

# Column (iv) number of 6-digit codes in matched data

# Importing dataset to get 6-digit industries
naics = pd.read_csv(int_dir / "main_annual.csv", usecols = ['gvkey','naics'])
naics = naics.drop_duplicates(subset = 'gvkey')

# Merging with the data for matched firms
df_matched = pd.merge(df_matched,naics,on='gvkey')
df_matched['naics'] = df_matched['naics'].astype(int)
df_matched['naics'] = df_matched['naics'].astype(str)

df_6digit = df_matched[['ind2d','naics']]
df_6digit = df_6digit.drop_duplicates(subset = 'naics')
df_6digit=df_6digit[df_6digit.naics.apply(lambda x: len(str(x))==6)]

n_6digit = df_6digit.pivot_table(columns=['ind2d'], aggfunc='size')
n_6digit = pd.DataFrame(n_6digit)                          
n_6digit.reset_index(inplace=True)
n_6digit.rename(columns={0:'N_6d_codes'}, inplace=True) 

stat_table = pd.merge(col123_t,n_6digit,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)
stat_table['N_6d_codes'] = stat_table['N_6d_codes'].astype(int)


# Column (v) average length of firm data

df_years = (df_matched.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d','ind2d_definition']]
        .agg(['first','last'])
        .reset_index())
df_years['periods'] = df_years['year','last'] - df_years['year','first'] + 1
df_years.columns = df_years.columns.map(','.join).str.strip(',')
df_years = df_years[['ind2d,first','periods']]
df_years.rename(columns={'ind2d,first':'ind2d'}, inplace=True)
mean_periods = df_years['periods'].mean()

df_n_years = df_years.groupby('ind2d').mean(numeric_only = True)
stat_table = pd.merge(stat_table,df_n_years,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)

# Adding industry definitions

defs = df[['ind2d','ind2d_definition']]
defs = defs.drop_duplicates(subset = 'ind2d')

stat_table = pd.merge(stat_table,defs,on='ind2d',how='outer')

stat_table.loc[0,'ind2d_definition'] = 'Total'
stat_table.loc[0,'N_6d_codes'] = stat_table.N_6d_codes[1:].sum()
stat_table.loc[0,'periods']= mean_periods
stat_table.loc[0,'ind2d'] = 'Total'

# Formatting the table

stat_table_annual = stat_table[['ind2d','ind2d_definition','N_firms',
                         'share_firms_matched','share_revenue_matched','N_6d_codes','periods']]
stat_table_annual.rename(columns={'ind2d':'Industry',
                          'ind2d_definition':'Definition',
                          'N_firms':'Number of firms',
                          'share_firms_matched': 'Share of firms matched',
                          'N_6d_codes':'Number of 6-digit industries',
                          'periods':'Number of periods (years)'},inplace=True)


# Exporting the table
stat_table_annual.round(4).to_csv(tab_dir / 'Stats Table Annual (truncated).csv')

# Saving as Tex file
with open(tab_dir / 'Stats Table Annual (truncated).tex','w') as tf:
    tf.write(stat_table_annual
             .rename(columns ={'share_revenue_matched': 'Share of revenue matched'})
             .round(2)
             .to_latex(index=False))






# ---------------------------------------------------------------------------------------- #
# For the quarterly data
# ---------------------------------------------------------------------------------------- #

# ## Generating table of statistics (Quarterly)

# Importing cleaned Compustat data

df_cpi = pd.read_csv(int_dir / "main_quarterly.csv")

df_cpi = df_cpi.dropna(subset='firm_level_markup')

# Normalizing PPI with CPI
df_cpi['PPI_CPI'] = df_cpi['ppi']/df_cpi['CPI']*100

main = df_cpi[['gvkey','year','quarter','cogsq','saleq','ind2d','firm_level_markup','ppi','CPI','ind2d_definition']].copy()

main.rename(columns={'saleq':'sale_raw',
                   'ppi':'PPI'}, inplace=True)
main = main[main.quarter != '2022Q4']
main.reset_index(drop=True, inplace=True)

# Deflating sales revenue using CPI
main['sale'] = (main['sale_raw']/main['CPI'])*100

# Editing the 2-digit industry codes
main['ind2d'] = main['ind2d'].astype(int)
main['ind2d'] = main['ind2d'].astype(str)
main['ind2d'] = main['ind2d'].astype(str)

# Manufacturing
main.loc[main.ind2d == '31', 'ind2d'] = '31-33'
main.loc[main.ind2d == '32', 'ind2d'] = '31-33'
main.loc[main.ind2d == '33', 'ind2d'] = '31-33'

# Retail trade
main.loc[main.ind2d == '44', 'ind2d'] = '44-45'
main.loc[main.ind2d == '45', 'ind2d'] = '44-45'

# Transportation and warehousing
main.loc[main.ind2d == '48', 'ind2d'] = '48-49'
main.loc[main.ind2d == '49', 'ind2d'] = '48-49'

# Adjusting the sample

# Length of firm data truncation

df_y = (main.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d']]
        .agg(['first','last'])
        .reset_index())
df_y['periods'] = df_y['year','last'] - df_y['year','first']
df_y.columns = df_y.columns.map(','.join).str.strip(',')
df_y = df_y[['gvkey','periods']]

# Combine length of firms data into the main dataframe
df = pd.merge(main,df_y,on='gvkey')
df = df[df.periods >= 4]
df = df[df.year >= 2018]
df.reset_index(drop=True, inplace=True)

# Including period index
index_quarter = pd.read_csv(data_dir / "Other" / "indexed_quarter.csv")
index_quarter.rename(columns={'index':'quarter_index'}, inplace=True)

# Merging with the data
df = pd.merge(df,index_quarter, on ='quarter')
df


# ## Data for all firms

# Column (i) Number of firms by 2-digit industries

firms = df[['ind2d','gvkey']]
firms = firms.drop_duplicates(subset = 'gvkey')

n_firms = firms.pivot_table(columns=['ind2d'], aggfunc='size')
n_firms = pd.DataFrame(n_firms)                          # To check the number of observations
n_firms.reset_index(inplace=True)
n_firms.rename(columns={0:'N_firms'}, inplace=True) 

print('Total number of firms:',n_firms.N_firms.sum())

# For Column (iii) summing up revenue for all firms

df_revenue = df.groupby('ind2d').sum(numeric_only = True)
df_revenue = df_revenue['sale']
df_revenue = pd.DataFrame(df_revenue)
df_revenue.reset_index(inplace=True)
df_revenue

col123 = pd.merge(n_firms, df_revenue, on='ind2d')


# ## Data for matched firms

# For Column (ii) share of all firms matched to PPI data

df_matched = main.dropna(subset=['PPI'])

# Dropping firms that have data for less than 5 years
df_y = (df_matched.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d']]
        .agg(['first','last'])
        .reset_index())
df_y['periods'] = df_y['year','last'] - df_y['year','first']
df_y.columns = df_y.columns.map(','.join).str.strip(',')
df_y = df_y[['gvkey','periods']]

# Combine length of firms data into the main dataframe
df_matched = pd.merge(df_matched,df_y,on='gvkey')
df_matched = df_matched[df_matched.periods >= 4]
#df_matched = df_matched[df_matched.year >= 2018]

# Adding index for quarters
df_matched = pd.merge(df_matched,index_quarter, on ='quarter')
df_matched.reset_index(drop=True, inplace=True)
df_matched

# Number of firms in 2-digit industries

firms_matched = df_matched[['ind2d','gvkey']]
firms_matched = firms_matched.drop_duplicates(subset = 'gvkey')

n_firms_matched = firms_matched.pivot_table(columns=['ind2d'], aggfunc='size')
n_firms_matched = pd.DataFrame(n_firms_matched)                          # To check the number of observations
n_firms_matched.reset_index(inplace=True)
n_firms_matched.rename(columns={0:'N_firms_m'}, inplace=True) 

print('Total number of matched firms:',n_firms_matched.N_firms_m.sum())


df_revenue_m = df_matched.groupby('ind2d').sum(numeric_only = True)
df_revenue_m = df_revenue_m['sale']
df_revenue_m = pd.DataFrame(df_revenue_m)
df_revenue_m.reset_index(inplace=True)
df_revenue_m.rename(columns={'sale':'sale_m'}, inplace=True) 

col123_m = pd.merge(n_firms_matched,df_revenue_m,on='ind2d')
col123_f = pd.merge(col123,col123_m,on='ind2d',how ='outer')
col123_f  = col123_f.fillna(0)


# Calculating total and adding as a row
col123_t = col123_f.sum(numeric_only=True)
col123_t = pd.DataFrame(col123_t).T
col123_t = pd.concat([col123_t,col123_f])
#col123_t = col123_t.append(col123_f, ignore_index=True)
col123_t.reset_index(drop=True, inplace=True)
col123_t

# Calculating values for Columns ii-iii
col123_t['share_firms_matched'] = col123_t['N_firms_m'] / col123_t['N_firms']
col123_t['share_revenue_matched'] = col123_t['sale_m'] / col123_t['sale']
col123_t = col123_t[['ind2d','N_firms','share_firms_matched','share_revenue_matched']].copy()
col123_t.reset_index(drop=True, inplace=True)


# ## Adding column (iv)

# Column (iv) number of 6-digit codes in matched data

# Importing main dataset to get 6-digit industries
naics = pd.read_csv(int_dir / "main_quarterly.csv", usecols = ['gvkey','naics'])
naics = naics.drop_duplicates(subset = 'gvkey')

# Merging with the data for matched firms
df_matched = pd.merge(df_matched,naics,on='gvkey')
df_matched['naics'] = df_matched['naics'].astype(int)
df_matched['naics'] = df_matched['naics'].astype(str)

# Keep only 6-digit industry codes
df_6digit = df_matched[['ind2d','naics']]
df_6digit = df_6digit.drop_duplicates(subset = 'naics')
df_6digit=df_6digit[df_6digit.naics.apply(lambda x: len(str(x))==6)]

n_6digit = df_6digit.pivot_table(columns=['ind2d'], aggfunc='size')
n_6digit = pd.DataFrame(n_6digit)                          # To check the number of observations
n_6digit.reset_index(inplace=True)
n_6digit.rename(columns={0:'N_6d_codes'}, inplace=True) 

stat_table = pd.merge(col123_t,n_6digit,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)
stat_table['N_6d_codes'] = stat_table['N_6d_codes'].astype(int)


# Column (v) average length of firm data

df_years = (df_matched.sort_values(['gvkey','quarter_index'])
        .groupby(['gvkey'])[['quarter_index','ind2d','ind2d_definition']]
        .agg(['first','last'])
        .reset_index())
df_years['periods'] = df_years['quarter_index','last'] - df_years['quarter_index','first'] + 1
df_years.columns = df_years.columns.map(','.join).str.strip(',')
df_years = df_years[['ind2d,first','periods']]
df_years.rename(columns={'ind2d,first':'ind2d'}, inplace=True)
mean_periods = df_years['periods'].mean()

df_n_years = df_years.groupby('ind2d').mean(numeric_only = True)
stat_table = pd.merge(stat_table,df_n_years,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)

# Adding industry definitions

defs = df[['ind2d','ind2d_definition']]
defs = defs.drop_duplicates(subset = 'ind2d')

stat_table = pd.merge(stat_table,defs,on='ind2d',how='outer')

stat_table.loc[0,'ind2d_definition'] = 'Total'
stat_table.loc[0,'N_6d_codes'] = stat_table.N_6d_codes[1:].sum()
stat_table.loc[0,'periods'] = mean_periods
stat_table.loc[0,'ind2d'] = 'Total'

# Formatting the table

stat_table_quarterly = stat_table[['ind2d','ind2d_definition','N_firms',
                         'share_firms_matched','share_revenue_matched','N_6d_codes','periods']]
stat_table_quarterly.rename(columns={'ind2d':'Industry',
                          'ind2d_definition':'Definition',
                          'N_firms':'Number of firms',
                          'share_firms_matched': 'Share of firms matched',
                          'N_6d_codes':'Number of 6-digit industries',
                          'periods':'Number of periods (quarters)'},inplace=True)

stat_table_quarterly


# Exporting the table

stat_table_quarterly.round(4).to_csv(tab_dir / 'Stats Table Quarterly.csv')

# Saving as Tex file
with open(tab_dir / 'Stats Table Quarterly.tex','w') as tf:
    tf.write(stat_table_quarterly
             .rename(columns ={'share_revenue_matched': 'Share of revenue matched'})
             .round(2)
             .to_latex(index=False))




# ## Table of statistics (quarterly, truncated)

# For Column (ii) share of all firms matched to PPI data

df_matched = main.dropna(subset=['PPI'])

# Dropping firms that have data for less than 5 years
df_y = (df_matched.sort_values(['gvkey','year'])
        .groupby(['gvkey'])[['year','ind2d']]
        .agg(['first','last'])
        .reset_index())
df_y['periods'] = df_y['year','last'] - df_y['year','first']
df_y.columns = df_y.columns.map(','.join).str.strip(',')
df_y = df_y[['gvkey','periods']]

# Combine length of firms data into the main dataframe
df_matched = pd.merge(df_matched,df_y,on='gvkey')
df_matched = df_matched[df_matched.periods >= 4]
#df_matched = df_matched[df_matched.year >= 2018]

# Adding index for quarters
df_matched = pd.merge(df_matched,index_quarter, on ='quarter')
df_matched.reset_index(drop=True, inplace=True)


# Importing CAGR data for truncation (following Figure 2)

cagr = pd.read_csv(int_dir / "scatter_quarterly.csv",usecols=['gvkey','cagr_PPI','cagr_markup'])
cagr = cagr.drop_duplicates()

df_matched = pd.merge(df_matched,cagr,on='gvkey')

# Applying Figure 2 truncations
# Truncating the data
df_matched = df_matched[df_matched.cagr_PPI <= 5]
df_matched = df_matched[df_matched.cagr_PPI >= -5]
df_matched = df_matched[df_matched.cagr_markup >= -30]
df_matched = df_matched[df_matched.cagr_markup <= 30]
df_matched


# Number of firms in 2-digit industries

firms_matched = df_matched[['ind2d','gvkey']]
firms_matched = firms_matched.drop_duplicates(subset = 'gvkey')

n_firms_matched = firms_matched.pivot_table(columns=['ind2d'], aggfunc='size')
n_firms_matched = pd.DataFrame(n_firms_matched)                          # To check the number of observations
n_firms_matched.reset_index(inplace=True)
n_firms_matched.rename(columns={0:'N_firms_m'}, inplace=True) 

print('Total number of matched firms:',n_firms_matched.N_firms_m.sum())


# Calculating revenue share

df_revenue_m = df_matched.groupby('ind2d').sum(numeric_only = True)
df_revenue_m = df_revenue_m['sale']
df_revenue_m = pd.DataFrame(df_revenue_m)
df_revenue_m.reset_index(inplace=True)
df_revenue_m.rename(columns={'sale':'sale_m'}, inplace=True) 

col123_m = pd.merge(n_firms_matched,df_revenue_m,on='ind2d')
col123_f = pd.merge(col123,col123_m,on='ind2d',how ='outer')
col123_f  = col123_f.fillna(0)


# Calculating total and adding as a row
col123_t = col123_f.sum(numeric_only=True)
col123_t = pd.DataFrame(col123_t).T
col123_t = pd.concat([col123_t,col123_f])
#col123_t = col123_t.append(col123_f, ignore_index=True)
col123_t.reset_index(drop=True, inplace=True)
col123_t

# Calculating values for Columns ii-iii
col123_t['share_firms_matched'] = col123_t['N_firms_m'] / col123_t['N_firms']
col123_t['share_revenue_matched'] = col123_t['sale_m'] / col123_t['sale']
col123_t = col123_t[['ind2d','N_firms','share_firms_matched','share_revenue_matched']].copy()
col123_t.reset_index(drop=True, inplace=True)


# Column (iv) number of 6-digit codes in matched data

# Importing main dataset to get 6-digit industries
naics = pd.read_csv(int_dir / "main_quarterly.csv", usecols = ['gvkey','naics'])
naics = naics.drop_duplicates(subset = 'gvkey')

# Merging with the data for matched firms
df_matched = pd.merge(df_matched,naics,on='gvkey')
df_matched['naics'] = df_matched['naics'].astype(int)
df_matched['naics'] = df_matched['naics'].astype(str)

# Keep only 6-digit industry codes
df_6digit = df_matched[['ind2d','naics']]
df_6digit = df_6digit.drop_duplicates(subset = 'naics')
df_6digit=df_6digit[df_6digit.naics.apply(lambda x: len(str(x))==6)]

n_6digit = df_6digit.pivot_table(columns=['ind2d'], aggfunc='size')
n_6digit = pd.DataFrame(n_6digit)                          # To check the number of observations
n_6digit.reset_index(inplace=True)
n_6digit.rename(columns={0:'N_6d_codes'}, inplace=True) 

stat_table = pd.merge(col123_t,n_6digit,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)
stat_table['N_6d_codes'] = stat_table['N_6d_codes'].astype(int)


# Column (v) average length of firm data

df_years = (df_matched.sort_values(['gvkey','quarter_index'])
        .groupby(['gvkey'])[['quarter_index','ind2d','ind2d_definition']]
        .agg(['first','last'])
        .reset_index())
df_years['periods'] = df_years['quarter_index','last'] - df_years['quarter_index','first'] + 1
df_years.columns = df_years.columns.map(','.join).str.strip(',')
df_years = df_years[['ind2d,first','periods']]
df_years.rename(columns={'ind2d,first':'ind2d'}, inplace=True)
mean_periods = df_years['periods'].mean()

df_n_years = df_years.groupby('ind2d').mean(numeric_only = True)
stat_table = pd.merge(stat_table,df_n_years,on='ind2d',how='outer')
stat_table  = stat_table.fillna(0)


# Adding industry definitions

defs = df[['ind2d','ind2d_definition']]
defs = defs.drop_duplicates(subset = 'ind2d')

stat_table = pd.merge(stat_table,defs,on='ind2d',how='outer')

stat_table.loc[0,'ind2d_definition'] = 'Total'
stat_table.loc[0,'N_6d_codes'] = stat_table.N_6d_codes[1:].sum()
stat_table.loc[0,'periods'] = mean_periods
stat_table.loc[0,'ind2d'] = 'Total'


# Formatting the table

stat_table_quarterly = stat_table[['ind2d','ind2d_definition','N_firms',
                         'share_firms_matched','share_revenue_matched','N_6d_codes','periods']]
stat_table_quarterly.rename(columns={'ind2d':'Industry',
                          'ind2d_definition':'Definition',
                          'N_firms':'Number of firms',
                          'share_firms_matched': 'Share of firms matched',
                          'N_6d_codes':'Number of 6-digit industries',
                          'periods':'Number of periods (quarters)'},inplace=True)

# Exporting the table

stat_table_quarterly.round(4).to_csv(tab_dir / 'Stats Table Quarterly (truncated).csv')

# Saving as Tex file
with open(tab_dir / 'Stats Table Quarterly (truncated).tex','w') as tf:
    tf.write(stat_table_quarterly
             .rename(columns ={'share_revenue_matched': 'Share of revenue matched'})
             .round(2)
             .to_latex(index=False))



"""
Shared helpers for the BMY replication pipeline.

These functions mirror common Stata idioms used in the original .do files:
- consistent paths rooted at this folder
- Stata .dta IO with pandas
- groupwise quantiles and lags
"""

from __future__ import annotations

from pathlib import Path
from typing import Iterable

import numpy as np
import pandas as pd


BASE_DIR = Path(__file__).resolve().parent
RAW_DATA_DIR = BASE_DIR / "rawData"
INTERMEDIATE_DIR = BASE_DIR / "intermediateOutput"
OUTPUT_FIGURES_DIR = BASE_DIR / "outputFigures"
TEMP_DIR = BASE_DIR / "temp"


def ensure_directories() -> None:
    """Create expected output folders if they do not exist."""
    for path in (INTERMEDIATE_DIR, OUTPUT_FIGURES_DIR, TEMP_DIR):
        path.mkdir(parents=True, exist_ok=True)


def read_dta(path: Path) -> pd.DataFrame:
    """Read a Stata file with pandas."""
    return pd.read_stata(path)


def write_dta(df: pd.DataFrame, path: Path) -> None:
    """Persist a DataFrame as a Stata .dta file."""
    path.parent.mkdir(parents=True, exist_ok=True)
    df.to_stata(path, write_index=False, version=118)


def group_quantile(df: pd.DataFrame, by: Iterable[str], col: str, q: float) -> pd.Series:
    """Groupwise quantile helper equivalent to Stata's pctile."""
    return df.groupby(list(by))[col].transform(lambda x: x.quantile(q))


def lag_by_group(df: pd.DataFrame, group_col: str, time_col: str, cols: Iterable[str], periods: int = 1) -> pd.DataFrame:
    """Add within-group lags for the requested columns."""
    df = df.sort_values([group_col, time_col]).copy()
    for col in cols:
        df[f"L.{col}"] = df.groupby(group_col)[col].shift(periods)
    return df


def safe_log(series: pd.Series) -> pd.Series:
    """Natural log that returns NaN for nonpositive inputs."""
    return pd.Series(np.where(series > 0, np.log(series), np.nan), index=series.index)




"""
Python replacement for ``0.4 Create Main Datasets.do``.

Builds the annual and quarterly main datasets, computes firm-level markups
using the estimated theta values, and exports the CSVs consumed by downstream
figure/table scripts.
"""

from __future__ import annotations

import argparse
import logging
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd

from Estimate_Coefficients import theta_suffix
from path_plot_config import data_dir, int_dir

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
LOGGER = logging.getLogger(__name__)


def _load_naics_desc(path: Path) -> pd.DataFrame:
    df = pd.read_excel(path)
    df.columns = df.columns.str.strip().str.lower()
    df = df.rename(columns={"sector_definition": "ind2d_definition"})
    df["ind2d"] = pd.to_numeric(df["ind2d"], errors="coerce")
    return df[["ind2d", "ind2d_definition"]]


def _load_ppi(path: Path, period_col: str) -> pd.DataFrame:
    df = pd.read_csv(path)
    df.columns = df.columns.str.strip().str.lower()
    df = df.rename(columns={"naics_code": "naics"})
    df["naics"] = df["naics"].astype(str).str.strip()
    df[period_col] = df[period_col].astype(str).str.strip()
    return df[[period_col, "naics", "ppi"]]


def _load_cpi(path: Path, period_col: str) -> pd.DataFrame:
    df = pd.read_csv(path)
    df.columns = df.columns.str.strip().str.lower()
    df[period_col] = df[period_col].astype(str).str.strip()
    return df[[period_col, "cpi"]]


def _apply_costshare_trim(df: pd.DataFrame, time_col: str) -> pd.DataFrame:
    df = df.copy()
    df["costshare0"] = 0.85
    df["costshare1"] = df["cogs_D"] / (df["cogs_D"] + df["kexp"])
    df["costshare2"] = df["cogs_D"] / (df["cogs_D"] + df["xsga_D"] + df["kexp"])
    for col in ("costshare1", "costshare2"):
        p1 = df.groupby(time_col)[col].transform(lambda x: x.quantile(0.01))
        p99 = df.groupby(time_col)[col].transform(lambda x: x.quantile(0.99))
        df = df[
            (df[col] > 0)
            & df[col].notna()
            & (df[col] > p1)
            & (df[col] < p99)
        ]
    return df


def _attach_theta(df: pd.DataFrame, theta_w_path: Path, theta_acf_path: Path) -> pd.DataFrame:
    if not theta_w_path.exists():
        raise FileNotFoundError(f"Missing theta file {theta_w_path}. Run 0.3 theta_estimation.py first.")
    theta = pd.read_stata(theta_w_path)
    theta.columns = theta.columns.str.lower()
    theta = theta.rename(columns={"theta_wi1_ct": "theta_WI1_ct"})
    df = df.merge(theta[["ind2d", "year", "theta_WI1_ct"]], on=["ind2d", "year"], how="left")

    if theta_acf_path.exists():
        theta_acf = pd.read_stata(theta_acf_path)
        theta_acf.columns = theta_acf.columns.str.lower()
        theta_acf = theta_acf.rename(columns={"theta_acf": "theta_acf_val"})
        df = df.merge(theta_acf[["ind2d", "year", "theta_acf_val"]], on=["ind2d", "year"], how="left")
    else:
        LOGGER.warning("theta_acf file missing (%s); OP/ACF markups will be NaN.", theta_acf_path)
        df["theta_acf_val"] = np.nan
    return df


def _compute_mu(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy()
    df["mu_10"] = df["theta_WI1_ct"] * (df["sale_D"] / df["cogs_D"])
    df["mu_12"] = df["theta_acf_val"] * (df["sale_D"] / df["cogs_D"])
    return df


def _aggregate_markup(df: pd.DataFrame, time_col: str, markup_col: str = "mu_10") -> pd.DataFrame:
    df = df.copy()
    df["TOTSALES"] = df.groupby(time_col)["sale_D"].transform("sum")
    df["share_firm_agg"] = df["sale_D"] / df["TOTSALES"]
    agg = (
        df.groupby(time_col)
        .apply(lambda g: np.sum(g["share_firm_agg"] * g[markup_col].fillna(0)))
        .reset_index(name="MARKUP_spec1")
    )
    return agg


def _save_for_figure1(df: pd.DataFrame, limited: bool, out_dir: Path, markup_col: str = "mu_10", label: str = "") -> None:
    filtered = df if not limited else df[df["ppi"].notna()]
    agg = _aggregate_markup(filtered, "year", markup_col=markup_col)
    out_dir.mkdir(parents=True, exist_ok=True)
    suffix = f"_{label}" if label else ""
    fname = ("agg_markup_limited_to_PPI matched_annual" + suffix + ".csv") if limited else ("agg_markup_annual" + suffix + ".csv")
    agg.to_csv(out_dir / fname, index=False)
    LOGGER.info("Saved %s", out_dir / fname)


def prepare_annual(data_root: Path, include_interest_cogs: bool, drop_missing_sga: bool) -> None:
    naics_desc = _load_naics_desc(data_root / "Input" / "Other" / "NAICS_2D_Description.xlsx")
    ppi = _load_ppi(data_root / "Input" / "PPI" / "PPI_annual.csv", "year")
    cpi = _load_cpi(data_root / "Input" / "CPI" / "CPI_annual.csv", "year")

    fname = "data_main_upd_trim_1_intExp.dta" if include_interest_cogs else "data_main_upd_trim_1.dta"
    main_path = data_root / "Intermediate" / fname
    if not main_path.exists():
        raise FileNotFoundError(f"Missing {main_path}. Run Create_Data.py first.")

    df = pd.read_stata(main_path)
    df["naics"] = df["naics"].astype(str).str.strip()
    df["year"] = pd.to_numeric(df["year"], errors="coerce")

    df = df.merge(ppi, on=["year", "naics"], how="left")
    df = df.merge(cpi, on="year", how="left")
    df = df.merge(naics_desc, on="ind2d", how="left")

    df = _apply_costshare_trim(df, time_col="year")
    df = df[df["year"] >= 1955]
    df = df[df["ind2d"] != 99]

    suffix = theta_suffix(include_interest_cogs, drop_missing_sga)
    theta_w_path = data_root / "Intermediate" / f"theta_W_s_window_{suffix}.dta"
    theta_acf_path = data_root / "Intermediate" / f"theta_acf_{suffix}.dta"
    df = _attach_theta(df, theta_w_path, theta_acf_path)
    df = _compute_mu(df)

    df = df.rename(columns={"mu_10": "firm_level_markup", "mu_12": "firm_level_markup_acf"})
    annual_out = df[
        ["gvkey", "conm", "year", "naics", "ind2d", "ind2d_definition", "firm_level_markup", "firm_level_markup_acf", "sale", "sale_D", "cogs", "cogs_D", "ppi", "cpi"]
    ].copy()
    int_dir.mkdir(parents=True, exist_ok=True)
    annual_out.to_csv(int_dir / "main_annual.csv", index=False)
    LOGGER.info("Saved %s", int_dir / "main_annual.csv")

    out_dir = data_root / "Intermediate" / "For Figure 1"
    _save_for_figure1(df, limited=False, out_dir=out_dir, markup_col="mu_10")
    _save_for_figure1(df, limited=True, out_dir=out_dir, markup_col="mu_10")
    # Optional OP/ACF aggregate outputs (won't affect plotting)
    _save_for_figure1(df, limited=False, out_dir=out_dir, markup_col="mu_12", label="op_acf")
    _save_for_figure1(df, limited=True, out_dir=out_dir, markup_col="mu_12", label="op_acf")


def prepare_quarterly(data_root: Path, include_interest_cogs: bool, drop_missing_sga: bool) -> None:
    naics_desc = _load_naics_desc(data_root / "Input" / "Other" / "NAICS_2D_Description.xlsx")
    ppi = _load_ppi(data_root / "Input" / "PPI" / "PPI_quarterly.csv", "quarter")
    cpi = _load_cpi(data_root / "Input" / "CPI" / "CPI_quarterly.csv", "quarter")
    macro = pd.read_excel(data_root / "Input" / "DLEU" / "macro_vars_new.xlsx")
    macro.columns = macro.columns.str.strip()
    macro = macro[["year", "USGDP", "usercost"]]

    comp_path = data_root / "Input" / "DLEU" / "Compustat_quarterly.dta"
    if not comp_path.exists():
        raise FileNotFoundError(f"Missing {comp_path}")
    df = pd.read_stata(comp_path)

    df = df[df["fqtr"].notna()]
    df["fyearq"] = df["fyearq"].astype(int).astype(str)
    df["fqtr"] = df["fqtr"].astype(int).astype(str)
    df["quarter"] = df["fyearq"] + "Q" + df["fqtr"]

    df = df.sort_values(["gvkey", "quarter"])
    df["nrobs"] = df.groupby(["gvkey", "quarter"])["quarter"].transform("size")
    df = df[~(((df["nrobs"] == 2) | (df["nrobs"] == 3)) & (df["indfmt"] == "FS"))]
    df = df.drop_duplicates(subset=["gvkey", "quarter"])

    df = df[df["naics"].notna() & (df["naics"] != "")]
    for digits in (2, 3, 4):
        df[f"ind{digits}d"] = pd.to_numeric(df["naics"].str.slice(0, digits), errors="coerce")

    keep_cols = ["gvkey", "fyearq", "quarter", "naics", "ind2d", "ind3d", "ind4d", "saleq", "cogsq", "xsgaq", "ppegtq", "conm"]
    df = df[keep_cols].copy()
    df[["saleq", "cogsq", "xsgaq", "ppegtq"]] = df[["saleq", "cogsq", "xsgaq", "ppegtq"]] * 1000

    df = df.rename(columns={"fyearq": "year"})
    df["year"] = pd.to_numeric(df["year"], errors="coerce")

    df = df.merge(macro, on="year", how="inner")
    df["naics"] = df["naics"].astype(str).str.strip()
    df = df.merge(ppi, on=["quarter", "naics"], how="left")
    df = df.merge(cpi, on="quarter", how="left")
    df = df.merge(naics_desc, on="ind2d", how="left")

    df["sale_D"] = (df["saleq"] / df["USGDP"]) * 100
    df["cogs_D"] = (df["cogsq"] / df["USGDP"]) * 100
    df["xsga_D"] = (df["xsgaq"] / df["USGDP"]) * 100
    df["capital_D"] = (df["ppegtq"] / df["USGDP"]) * 100
    df["kexp"] = df["usercost"] * df["capital_D"]

    df = df[df["sale_D"] >= 0]
    df = df[df["cogs_D"] >= 0]
    df = df[df["xsgaq"] >= 0]

    df["s_g"] = df["saleq"] / df["cogsq"]
    p1 = df.groupby("quarter")["s_g"].transform(lambda x: x.quantile(0.01))
    p99 = df.groupby("quarter")["s_g"].transform(lambda x: x.quantile(0.99))
    df = df[(df["s_g"] > p1) & (df["s_g"] < p99)]
    df = df.drop(columns=["s_g"])

    df = _apply_costshare_trim(df, time_col="quarter")
    suffix = theta_suffix(include_interest_cogs, drop_missing_sga)
    theta_w_path = data_root / "Intermediate" / f"theta_W_s_window_{suffix}.dta"
    theta_acf_path = data_root / "Intermediate" / f"theta_acf_{suffix}.dta"
    df = _attach_theta(df, theta_w_path, theta_acf_path)
    df = _compute_mu(df)

    df = df[df["year"] >= 1955]
    df = df[df["ind2d"] != 99]
    df = df.rename(columns={"mu_10": "firm_level_markup", "mu_12": "firm_level_markup_acf"})

    quarterly_out = df[
        ["gvkey", "conm", "year", "quarter", "naics", "ind2d", "ind2d_definition", "firm_level_markup", "firm_level_markup_acf", "saleq", "sale_D", "cogsq", "cogs_D", "ppi", "cpi"]
    ].copy()
    quarterly_out.to_csv(int_dir / "main_quarterly.csv", index=False)
    LOGGER.info("Saved %s", int_dir / "main_quarterly.csv")


def main(
    data_root: Optional[Path] = None,
    include_interest_cogs: bool = False,
    drop_missing_sga: bool = False,
) -> None:
    root = data_root or Path(__file__).resolve().parents[2]
    prepare_annual(root, include_interest_cogs=include_interest_cogs, drop_missing_sga=drop_missing_sga)
    prepare_quarterly(root, include_interest_cogs=include_interest_cogs, drop_missing_sga=drop_missing_sga)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Python replacement for 0.4 Create Main Datasets.do")
    parser.add_argument("--data-root", type=Path, default=None, help="Project root containing Input/ and Intermediate/.")
    parser.add_argument("--include-interest-cogs", type=int, choices=[0, 1], default=0)
    parser.add_argument("--drop-missing-sga", type=int, choices=[0, 1], default=0)
    args = parser.parse_args()

    main(
        data_root=args.data_root,
        include_interest_cogs=bool(args.include_interest_cogs),
        drop_missing_sga=bool(args.drop_missing_sga),
    )



"""Wooldridge-style IV/GMM estimator for production function elasticities."""

from __future__ import annotations

import logging
from typing import Literal

import numpy as np
import pandas as pd
from linearmodels.iv import IV2SLS

from PyMarkup.core.data_preparation import add_lags, safe_log
from PyMarkup.estimators.base import ProductionFunctionEstimator

logger = logging.getLogger(__name__)


class WooldridgeIVEstimator(ProductionFunctionEstimator):
    """
    Wooldridge-style IV/GMM estimator using lagged COGS as instrument.

    This estimator addresses simultaneity bias in production function estimation
    by instrumenting current COGS with lagged COGS, following Wooldridge's approach.

    Two specifications are available:
    - spec1: Output ~ COGS + Capital (+ controls)
    - spec2: Output ~ COGS + Capital + SG&A (+ controls)

    Parameters
    ----------
    specification : {"spec1", "spec2", "both"}
        Which specification to estimate:
        - "spec1": COGS + capital only
        - "spec2": COGS + capital + SG&A
        - "both": estimate both, return spec2 results
    window_years : int
        Rolling window size in years (default: 5 = ±2 years)
    industry_level : int
        NAICS digit level for industry grouping (2, 3, or 4)
    min_observations : int
        Minimum observations required per window (default: 15)

    Attributes
    ----------
    results_ : pd.DataFrame
        Estimation results after calling estimate_elasticities()

    Examples
    --------
    >>> estimator = WooldridgeIVEstimator(specification="spec2", window_years=5)
    >>> elasticities = estimator.estimate_elasticities(panel_data)
    >>> print(elasticities.head())
    """

    def __init__(
        self,
        specification: Literal["spec1", "spec2", "both"] = "spec2",
        window_years: int = 5,
        industry_level: int = 2,
        min_observations: int = 15,
    ):
        if industry_level not in {2, 3, 4}:
            raise ValueError(f"industry_level must be 2, 3, or 4, got {industry_level}")

        self.specification = specification
        self.window_years = window_years
        self.industry_level = industry_level
        self.min_observations = min_observations
        self.results_ = None

    def get_method_name(self) -> str:
        """Return method name."""
        return f"Wooldridge IV ({self.specification})"

    def _preprocess(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Prepare data for IV estimation.

        Creates log-transformed variables, polynomials, and lags.
        """
        df = data.copy()

        # Create firm ID if not exists
        if "id" not in df.columns:
            df["id"] = df["gvkey"].astype("category").cat.codes

        # Log-transform main variables
        df["r"] = np.log(df["sale_D"])  # Output (revenue)
        df["y"] = df["r"]
        df["c"] = np.log(df["cogs_D"])  # Variable input
        df["k"] = np.log(df["capital_D"])  # Capital

        # Polynomials
        df["c2"] = df["c"] ** 2
        df["c3"] = df["c"] ** 3
        df["k2"] = df["k"] ** 2
        df["k3"] = df["k"] ** 3
        df["ck"] = df["c"] * df["k"]

        # SG&A (if using spec2)
        df["lsga"] = safe_log(df["xsga_D"])
        df["lsga2"] = df["lsga"] ** 2

        # Capital stock and investment
        df["K"] = np.exp(df["k"])
        df["depr"] = 0.1
        df = add_lags(df, group="id", time="year", cols=["K"])
        df["Inv"] = df["K"] - (1 - df["depr"]) * df["L.K"]
        df["i"] = safe_log(df["Inv"])
        df["i2"] = df["i"] ** 2

        # Lags for instruments and controls
        df = add_lags(df, group="id", time="year", cols=["c", "k", "i", "lsga", "k2", "lsga2"])

        return df

    def _run_iv(
        self,
        df: pd.DataFrame,
        dep: str,
        endog: str,
        exog: list[str],
        instruments: list[str],
    ) -> dict[str, float] | None:
        """
        Run IV/2SLS regression.

        Parameters
        ----------
        df : pd.DataFrame
            Data for estimation
        dep : str
            Dependent variable
        endog : str
            Endogenous variable
        exog : list[str]
            Exogenous variables (controls)
        instruments : list[str]
            Instrumental variables

        Returns
        -------
        dict or None
            Dictionary of coefficients, or None if estimation fails
        """
        cols_needed = [dep, endog, *exog, *instruments]
        work = df[cols_needed].dropna()

        if work.shape[0] < self.min_observations:
            return None

        try:
            model = IV2SLS(
                dependent=work[dep],
                exog=pd.concat([pd.Series(1.0, index=work.index, name="const"), work[exog]], axis=1),
                endog=work[[endog]],
                instruments=work[instruments],
            )
            res = model.fit(cov_type="robust")
            return res.params.to_dict()
        except Exception as exc:
            logger.warning(f"IV estimation failed: {exc}")
            return None

    def _estimate_window(self, df: pd.DataFrame) -> tuple[dict | None, dict | None]:
        """
        Estimate both specifications for a given window.

        Returns
        -------
        tuple
            (spec1_coefficients, spec2_coefficients)
        """
        # Specification 1: COGS + Capital
        spec1 = self._run_iv(
            df,
            dep="r",
            endog="c",
            exog=["k", "L.i", "L.k2", "L.k"],
            instruments=["L.c"],
        )

        # Specification 2: COGS + Capital + SG&A
        spec2 = self._run_iv(
            df,
            dep="r",
            endog="c",
            exog=["k", "lsga", "L.i", "L.k2", "L.lsga2", "L.k", "L.lsga"],
            instruments=["L.c"],
        )

        return spec1, spec2

    def _build_result_frame(self, df: pd.DataFrame) -> pd.DataFrame:
        """Create container for results."""
        ind_col = f"ind{self.industry_level}d"
        ind_num_col = f"nrind{self.industry_level}"

        # Get unique industry-year combinations
        res = df[[ind_col, "year"]].drop_duplicates().copy()

        # Store actual industry code for output (don't rename yet)
        res["ind2d"] = res[ind_col]

        # Initialize coefficient columns
        for col in ["theta_WI1_ct", "theta_WI2_ct", "theta_WI2_xt", "theta_WI1_kt", "theta_WI2_kt"]:
            res[col] = np.nan

        return res

    def _assign_coefficients(
        self,
        res: pd.DataFrame,
        condition: pd.Series,
        spec1: dict | None,
        spec2: dict | None,
    ) -> None:
        """Assign estimated coefficients to result frame."""
        if spec1:
            if "c" in spec1:
                res.loc[condition, "theta_WI1_ct"] = spec1["c"]
            if "k" in spec1:
                res.loc[condition, "theta_WI1_kt"] = spec1["k"]

        if spec2:
            if "c" in spec2:
                res.loc[condition, "theta_WI2_ct"] = spec2["c"]
            if "lsga" in spec2:
                res.loc[condition, "theta_WI2_xt"] = spec2["lsga"]
            if "k" in spec2:
                res.loc[condition, "theta_WI2_kt"] = spec2["k"]

    def _run_rolling_windows(self, df: pd.DataFrame, res: pd.DataFrame) -> None:
        """
        Run rolling window estimation for all industries and years.

        This implements the full logic from the original Stata code, including:
        - Early windows for industries 1-16, 18-25
        - Special handling for industry 17
        - Rolling 5-year windows for all industries from 1970 onwards
        """
        ind_col = f"nrind{self.industry_level}"

        # Helper function to estimate for a subset
        def estimate_for_subset(subset: pd.DataFrame, assign_mask: pd.Series) -> None:
            spec1, spec2 = self._estimate_window(subset)
            if spec1 or spec2:
                self._assign_coefficients(res, assign_mask, spec1, spec2)

        # Early windows: industries 1-16 (before 1970)
        for s in range(1, 17):
            subset = df[
                (df[ind_col] == s) & (df["year"] < 1972) & df[["r", "c", "k", "L.c", "L.i", "L.k"]].notna().all(axis=1)
            ]
            if subset.shape[0] > self.min_observations:
                mask = (res["ind2d"] == s) & (res["year"] < 1970)
                estimate_for_subset(subset, mask)

        # Early windows: industries 18-25 (before 1970)
        for s in range(18, 26):
            subset = df[
                (df[ind_col] == s) & (df["year"] < 1972) & df[["r", "c", "k", "L.c", "L.i", "L.k"]].notna().all(axis=1)
            ]
            if subset.shape[0] > self.min_observations:
                mask = (res["ind2d"] == s) & (res["year"] < 1970)
                estimate_for_subset(subset, mask)

        # Special industry 17 (before 1985)
        subset_17 = df[
            (df[ind_col] == 17) & (df["year"] < 1985) & df[["r", "c", "k", "L.c", "L.i", "L.k"]].notna().all(axis=1)
        ]
        if subset_17.shape[0] > self.min_observations:
            mask = (res["ind2d"] == 17) & (res["year"] < 1985)
            estimate_for_subset(subset_17, mask)

        # Rolling windows from 1970 onwards
        window_half = self.window_years // 2

        for t in range(1970, 2025):
            window_mask = (df["year"] >= t - window_half) & (df["year"] <= t + window_half)

            # Industries 1-16
            for s in range(1, 17):
                subset = df[
                    (df[ind_col] == s)
                    & window_mask
                    & df[["r", "c", "k", "L.c", "L.i", "L.k"]].notna().all(axis=1)
                ]
                if subset.shape[0] > self.min_observations:
                    mask = (res["ind2d"] == s) & (res["year"] == t)
                    estimate_for_subset(subset, mask)

            # Industries 18-25
            for s in range(18, 26):
                subset = df[
                    (df[ind_col] == s)
                    & window_mask
                    & df[["r", "c", "k", "L.c", "L.i", "L.k"]].notna().all(axis=1)
                ]
                if subset.shape[0] > self.min_observations:
                    mask = (res["ind2d"] == s) & (res["year"] == t)
                    estimate_for_subset(subset, mask)

            # Industry 17 (from 1985 onwards)
            if t >= 1985:
                subset = df[
                    (df[ind_col] == 17)
                    & window_mask
                    & df[["r", "c", "k", "L.c", "L.i", "L.k"]].notna().all(axis=1)
                ]
                if subset.shape[0] > self.min_observations:
                    mask = (res["ind2d"] == 17) & (res["year"] == t)
                    estimate_for_subset(subset, mask)

    def estimate_elasticities(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Estimate output elasticities using rolling windows by industry.

        Parameters
        ----------
        data : pd.DataFrame
            Prepared panel data with required columns:
            - gvkey, year, sale_D, cogs_D, capital_D, xsga_D (optional)
            - ind2d, nrind2 (or ind3d/nrind3, ind4d/nrind4)

        Returns
        -------
        pd.DataFrame
            Elasticity estimates with columns:
            - ind2d: industry code
            - year: fiscal year
            - theta_c: COGS elasticity (from chosen specification)
            - theta_k: capital elasticity
        """
        logger.info(f"Starting {self.get_method_name()} estimation")
        logger.info(f"Window size: {self.window_years} years, Industry level: {self.industry_level}-digit NAICS")

        # Preprocess data
        df = self._preprocess(data)

        # Build result container
        res = self._build_result_frame(df)

        # Run rolling window estimation
        logger.info("Running rolling window IV estimation...")
        self._run_rolling_windows(df, res)

        # Count successful estimates
        n_estimates = res["theta_WI1_ct"].notna().sum() if self.specification == "spec1" else res["theta_WI2_ct"].notna().sum()
        logger.info(f"Successfully estimated {n_estimates} industry-year elasticities")

        # Select output columns based on specification
        if self.specification == "spec1":
            res["theta_c"] = res["theta_WI1_ct"]
            res["theta_k"] = res["theta_WI1_kt"]
        elif self.specification == "spec2":
            res["theta_c"] = res["theta_WI2_ct"]
            res["theta_k"] = res["theta_WI2_kt"]
        else:  # both
            # Return spec2 by default (more controls)
            res["theta_c"] = res["theta_WI2_ct"]
            res["theta_k"] = res["theta_WI2_kt"]

        # Clean output
        output = res[["ind2d", "year", "theta_c", "theta_k"]].copy()
        output = output.dropna(subset=["theta_c"])  # Drop rows with no estimates

        self.results_ = output
        return output
